"""Exceptions occurring when talking to the devolo Home Control central unit."""


class GatewayOfflineError(Exception):
    """The requested gateway is offline and cannot be used."""

    def __init__(self) -> None:
        """Initialize error."""
        super().__init__("Gateway is offline.")



import json
import pathlib

import pytest

file = pathlib.Path(__file__).parent / "test_data.json"
with file.open("r") as fh:
    test_data = json.load(fh)

pytest_plugins = [
    "tests.fixtures.gateway",
    "tests.fixtures.homecontrol",
    "tests.fixtures.mprm",
    "tests.fixtures.mydevolo",
    "tests.fixtures.publisher",
    "tests.fixtures.requests",
    "tests.fixtures.socket",
    "tests.fixtures.updater",
]


@pytest.fixture(autouse=True)
def test_data_fixture(request):
    """Load test data."""
    request.cls.user = test_data["user"]
    request.cls.gateway = test_data["gateway"]
    request.cls.devices = test_data["devices"]



import pytest

from ..mocks.mock_response import (
    MockResponseConnectTimeout,
    MockResponseGet,
    MockResponseJsonError,
    MockResponsePost,
    MockResponseReadTimeout,
    MockResponseServiceUnavailable,
)


@pytest.fixture()
def mock_response_gateway_offline(mocker):
    """Mock requests get method with service_unavailable status_code."""
    mocker.patch("requests.get", return_value=MockResponseServiceUnavailable({"link": "test_link"}, status_code=503))


@pytest.fixture()
def mock_response_json(mocker):
    """Mock requests.Session get method with success status_code."""
    mocker.patch("requests.Session", return_value=MockResponseGet({"link": "test_link"}, status_code=200))


@pytest.fixture()
def mock_response_requests_ConnectTimeout(mocker):
    """Mock requests.Session with ConnectTimeout exception."""
    mocker.patch("requests.Session", return_value=MockResponseConnectTimeout({"link": "test_link"}, status_code=200))


@pytest.fixture()
def mock_response_json_JSONDecodeError(mocker):
    """Mock requests.Session with JSONDecodeError exception."""
    mocker.patch("requests.Session", return_value=MockResponseJsonError({"link": "test_link"}, status_code=200))


@pytest.fixture()
def mock_response_requests_invalid_id(mocker):
    """Mock requests.Session with JSONDecodeError exception."""
    mocker.patch("requests.Session", return_value=MockResponsePost({"link": "test_link"}, status_code=200))


@pytest.fixture()
def mock_response_requests_valid(mocker):
    """Mock requests.Session post method with success status_code."""
    mocker.patch("requests.Session", return_value=MockResponsePost({"link": "test_link"}, status_code=200))


@pytest.fixture()
def mock_response_valid(mocker):
    """Mock requests get method with success status_code."""
    mocker.patch("requests.get", return_value=MockResponseGet({"response": "response"}, status_code=200))


@pytest.fixture()
def mock_response_wrong_credentials_error(mocker):
    """Mock requests get method with forbidden status_code."""
    mocker.patch("requests.get", return_value=MockResponseGet({"link": "test_link"}, status_code=403))


@pytest.fixture()
def mock_response_wrong_url_error(mocker):
    """Mock requests get method with notfound status_code."""
    mocker.patch("requests.get", return_value=MockResponseGet({"link": "test_link"}, status_code=404))


@pytest.fixture()
def mock_response_requests_ReadTimeout(mocker):
    """Mock requests.Session with ReadTimeout exception."""
    mocker.patch("requests.Session", return_value=MockResponseReadTimeout({"link": "test_link"}, status_code=200))


@pytest.fixture()
def mock_session_get(mocker, request):
    """Mock requests.Session get method with test data."""
    properties = {
        "test_get_local_session_valid": {"link": "test_link"},
        "test__on_pong": {"link": "test_link"},
    }
    mocker.patch("requests.Session", return_value=MockResponseGet(properties.get(request.node.name), status_code=200))


@pytest.fixture()
def mock_session_get_nok(mocker):
    mocker.patch("requests.Session", return_value=MockResponseServiceUnavailable({}, status_code=500))



"""Fixtures used for mydevolo testing."""
from typing import Generator
from unittest.mock import patch

import pytest
from pytest_mock import MockerFixture

from devolo_home_control_api.mydevolo import Mydevolo, WrongCredentialsError, WrongUrlError

from ..mocks.mock_mydevolo import MockMydevolo


@pytest.fixture()
def mydevolo(request: pytest.FixtureRequest) -> Generator[Mydevolo, None, None]:
    """Create real mydevolo object with static test data."""
    with patch("devolo_home_control_api.mydevolo.Mydevolo.uuid", return_value=request.cls.user.get("uuid")):
        yield Mydevolo()


@pytest.fixture()
def mock_mydevolo__call(mocker: MockerFixture, request: pytest.FixtureRequest) -> None:
    """Mock calls to the mydevolo API."""
    mock_mydevolo = MockMydevolo(request)
    mocker.patch("devolo_home_control_api.mydevolo.Mydevolo._call", side_effect=mock_mydevolo._call)


@pytest.fixture()
def mock_mydevolo__call_raise_WrongUrlError(mocker: MockerFixture) -> None:
    """Respond with WrongUrlError on calls to the mydevolo API."""
    mocker.patch("devolo_home_control_api.mydevolo.Mydevolo._call", side_effect=WrongUrlError)


@pytest.fixture()
def mock_mydevolo_uuid_raise_WrongCredentialsError(mydevolo: Mydevolo) -> Generator[Mydevolo, None, None]:
    """Respond with WrongCredentialsError on calls to the mydevolo API."""
    with patch("devolo_home_control_api.mydevolo.Mydevolo.uuid", side_effect=WrongCredentialsError):
        yield mydevolo


@pytest.fixture()
def mock_get_zwave_products(mocker: MockerFixture) -> None:
    """Mock Z-Wave product information call to speed up tests."""
    mocker.patch("devolo_home_control_api.mydevolo.Mydevolo.get_zwave_products", return_value={})



"""REST communication"""
import json
import logging
import sys
from abc import ABC
from typing import Any, Dict, List, Union

from requests import Session
from requests.exceptions import ConnectionError, ReadTimeout  # pylint: disable=redefined-builtin

from ..devices.gateway import Gateway
from ..exceptions.gateway import GatewayOfflineError
from ..mydevolo import Mydevolo


class MprmRest(ABC):
    """
    The abstract MprmRest object handles calls to the so called mPRM. It does not cover all API calls, just those requested
    up to now. All calls are done in a gateway context, so you have to create a derived class, that provides a Gateway object
    and a Session object.
    """

    def __init__(self) -> None:
        logging.captureWarnings(True)

        self._logger = logging.getLogger(self.__class__.__name__)
        self._data_id = 0
        self._local_ip = ""
        self._url = ""

        self._mydevolo: Mydevolo
        self._session: Session
        self.gateway: Gateway

    def get_all_devices(self) -> List[str]:
        """
        Get all devices.

        :return: All devices and their properties.
        """
        self._logger.info("Inspecting devices")
        data = {"method": "FIM/getFunctionalItems", "params": [["devolo.DevicesPage"], 0]}
        response = self._post(data)
        self._logger.debug("Response of 'get_all_devices':\n%s", response)
        return response["result"]["items"][0]["properties"]["deviceUIDs"]

    def get_all_zones(self) -> Dict[str, str]:
        """
        Get all zones, also called rooms.

        :return: All zone IDs and their name.
        """
        self._logger.debug("Inspecting zones")
        data = {"method": "FIM/getFunctionalItems", "params": [["devolo.Grouping"], 0]}
        response = self._post(data)["result"]["items"][0]["properties"]["zones"]
        self._logger.debug("Response of 'get_all_zones':\n%s", response)
        return {key["id"]: key["name"] for key in response}

    def get_data_from_uid_list(self, uids: List[str]) -> List[Dict[str, Any]]:
        """
        Returns data from an element UID list using an RPC call.

        :param uids: Element UIDs, something like [devolo.MultiLevelSensor:hdm:ZWave:CBC56091/24#2,
                     devolo.MultiLevelSensor:hdm:ZWave:CBC56091/24#1]
        :return: Data connected to the element UIDs, payload so to say
        """
        data = {"method": "FIM/getFunctionalItems", "params": [uids, 0]}
        response = self._post(data)
        self._logger.debug("Response of 'get_data_from_uid_list':\n%s", response)
        return response["result"]["items"]

    def get_name_and_element_uids(self, uid: str) -> Dict[str, Any]:
        """
        Returns the name, all element UIDs and the device model of the given device UID.

        :param uid: Element UID, something like devolo.MultiLevelSensor:hdm:ZWave:CBC56091/24#2
        """
        data = {"method": "FIM/getFunctionalItems", "params": [[uid], 0]}
        response = self._post(data)
        self._logger.debug("Response of 'get_name_and_element_uids':\n%s", response)
        return response["result"]["items"][0]["properties"]

    def refresh_session(self):
        """
        Refresh currently running session. Without this call from time to time especially websockets will terminate.
        """
        self._logger.debug("Refreshing session.")
        data = {
            "method": "FIM/invokeOperation",
            "params": [f"devolo.UserPrefs.{self._mydevolo.uuid()}", "resetSessionTimeout", []],
        }
        self._post(data)

    def set_binary_switch(self, uid: str, state: bool) -> bool:
        """
        Set a binary switch state of a device.

        :param uid: Element UID, something like devolo.BinarySwitch:hdm:ZWave:CBC56091/24
        :param state: True if switching on, False if switching off
        :return: True if successfully switched, false otherwise
        """
        data = {"method": "FIM/invokeOperation", "params": [uid, "turnOn" if state else "turnOff", []]}
        response = self._post(data)
        return self._evaluate_response(uid=uid, value=state, response=response)

    def set_multi_level_switch(self, uid: str, value: float) -> bool:
        """
        Set a multi level switch value of a device.

        :param uid: Element UID, something like devolo.Dimmer:hdm:ZWave:CBC56091/24
        :param value: Value the multi level switch shall have
        :return: True if successfully switched, false otherwise
        """
        data = {"method": "FIM/invokeOperation", "params": [uid, "sendValue", [value]]}
        response = self._post(data)
        return self._evaluate_response(uid=uid, value=value, response=response)

    def set_remote_control(self, uid: str, key_pressed: int) -> bool:
        """
        Press the button of a remote control virtually.

        :param uid: Element UID, something like devolo.RemoteControl:hdm:ZWave:CBC56091/24
        :param key_pressed: Number of the button pressed
        :return: True if successfully switched, false otherwise
        """
        data = {"method": "FIM/invokeOperation", "params": [uid, "pressKey", [key_pressed]]}
        response = self._post(data)
        return self._evaluate_response(uid=uid, value=key_pressed, response=response)

    def set_setting(self, uid: str, setting: List[str]) -> bool:
        """
        Set a setting of a device.

        :param uid: Element UID, something like acs.hdm:ZWave:CBC56091/24
        :param setting: Settings to set
        :return: True if successfully switched, false otherwise
        """
        data = {"method": "FIM/invokeOperation", "params": [uid, "save", setting]}
        response = self._post(data)
        return self._evaluate_response(uid=uid, value=setting, response=response)

    def _evaluate_response(self, uid: str, value: Union[bool, float, List[str]], response: Dict[str, Any]) -> bool:
        """Evaluate the response of setting a device to a value."""
        if response["result"].get("status") == 1:
            return True
        if response["result"].get("status") == 2:
            self._logger.debug("Value of %s is already %s.", uid, value)
        else:
            self._logger.error("Something went wrong setting %s.", uid)
            self._logger.debug("Response to set command:\n%s", response)
        return False

    def _post(self, data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Communicate with the RPC interface. If the call times out, it is assumed that the gateway is offline and the state is
        changed accordingly.

        :param data: Data to be send
        :return: Response to the data
        """
        self._data_id += 1
        data["jsonrpc"] = "2.0"
        data["id"] = self._data_id
        try:
            response = self._session.post(
                f"{self._url}/remote/json-rpc", data=json.dumps(data), headers={"content-type": "application/json"}, timeout=30
            ).json()

        except (ConnectionError, ReadTimeout):
            self._logger.error("Gateway is offline.")
            self._logger.debug(sys.exc_info())
            self.gateway.update_state(False)
            raise GatewayOfflineError from None
        if response["id"] != data["id"]:
            self._logger.error("Got an unexpected response after posting data.")
            self._logger.debug("Message had ID %s, response had ID %s.", data["id"], response["id"])
            raise ValueError("Got an unexpected response after posting data.")
        return response



"""The Publisher"""
import logging
from typing import Any, Callable, Dict, KeysView, List, Optional, Tuple, Union


class Publisher:
    """
    The Publisher send messages to attached subscribers.
    """

    def __init__(self, events: Union[List[Any], KeysView]) -> None:
        self._logger = logging.getLogger(self.__class__.__name__)
        self._events: Dict[Any, Any] = {event: {} for event in events}

    def add_event(self, event: str) -> None:
        """Add a new event to listen to."""
        self._events[event] = {}

    def delete_event(self, event: str) -> None:
        """Delete a not longer needed event."""
        self._events.pop(event)

    def dispatch(self, event: str, message: Tuple[Any, ...]) -> None:
        """Dispatch the message to the subscribers."""
        for callback in self._get_subscribers_for_specific_event(event).values():
            callback(message)

    def register(self, event: str, who: Any, callback: Optional[Callable] = None) -> None:
        """
        As a new subscriber for an event, add a callback function to call on new message.
        If no callback is given, it registers update().

        :raises AttributeError: The supposed callback is not callable.
        """
        if callback is None:
            callback = getattr(who, "update")
        self._get_subscribers_for_specific_event(event)[who] = callback
        self._logger.debug("Subscriber registered for event %s", event)

    def unregister(self, event: str, who: Any) -> None:
        """Remove a subscriber for a specific event."""
        del self._get_subscribers_for_specific_event(event)[who]
        self._logger.debug("Subscriber deleted for event %s", event)

    def _get_subscribers_for_specific_event(self, event: str) -> Dict[Any, Any]:
        """All subscribers listening to an event."""
        return self._events.get(event, {})



import json
import pathlib

from devolo_home_control_api.devices.zwave import Zwave
from devolo_home_control_api.mydevolo import Mydevolo
from devolo_home_control_api.properties.multi_level_switch_property import MultiLevelSwitchProperty
from devolo_home_control_api.properties.settings_property import SettingsProperty


def siren(device_uid: str) -> Zwave:
    """
    Represent a siren in tests

    :param device_uid: Device UID this mock shall have
    :return: Siren device
    """
    file = pathlib.Path(__file__).parent / ".." / "test_data.json"
    with file.open("r") as fh:
        test_data = json.load(fh)

    mydevolo = Mydevolo()
    device = Zwave(mydevolo_instance=mydevolo, **test_data["devices"]["siren"])

    device.multi_level_switch_property = {}
    device.multi_level_switch_property[f"devolo.SirenMultiLevelSwitch:{device_uid}"] = MultiLevelSwitchProperty(
        element_uid=f"devolo.SirenMultiLevelSwitch:{device_uid}",
        setter=lambda uid, state: None,
        state=test_data["devices"]["siren"]["state"],
    )

    device.settings_property = {}
    device.settings_property["muted"] = SettingsProperty(
        element_uid=f"bas.{device_uid}", setter=lambda uid, state: True, value=test_data["devices"]["siren"]["muted"]
    )

    device.settings_property["general_device_settings"] = SettingsProperty(
        element_uid=f"gds.{device_uid}",
        setter=lambda uid, state: None,
        icon=test_data["devices"]["siren"]["icon"],
        name=test_data["devices"]["siren"]["itemName"],
        zone_id=test_data["devices"]["siren"]["zoneId"],
        zones=test_data["gateway"]["zones"],
    )

    device.settings_property["tone"] = SettingsProperty(
        element_uid=f"mss.{device_uid}",
        setter=lambda uid, state: None,
        value=test_data["devices"]["siren"]["properties"]["value"],
    )

    return device



"""Test handling the gateway device."""
from unittest.mock import patch

import pytest

from devolo_home_control_api.devices.gateway import Gateway
from devolo_home_control_api.exceptions.gateway import GatewayOfflineError
from devolo_home_control_api.mydevolo import Mydevolo


@pytest.mark.usefixtures("mock_mydevolo__call")
class TestGateway:
    def test_offline_init(self, mydevolo: Mydevolo) -> None:
        """Test, that a gateway can be setup while being offline."""
        with patch("devolo_home_control_api.mydevolo.Mydevolo.get_full_url", side_effect=GatewayOfflineError):
            gateway = Gateway(self.gateway.get("id"), mydevolo_instance=mydevolo)
            assert not gateway.full_url

    def test_update_state_known(self, mydevolo: Mydevolo) -> None:
        """Test state change to knowingly being offline."""
        gateway = Gateway(self.gateway.get("id"), mydevolo_instance=mydevolo)
        gateway.update_state(False)
        assert not gateway.online
        assert not gateway.sync

    def test_update_state_offline(self, mydevolo: Mydevolo) -> None:
        """Test state change to offline because of an update."""
        gateway = Gateway(self.gateway.get("id"), mydevolo_instance=mydevolo)
        gateway._update_state(status="devolo.hc_gateway.status.offline", state="devolo.hc_gateway.state.update")
        assert not gateway.online
        assert not gateway.sync

    def test_update_state_unknown(self, mydevolo: Mydevolo) -> None:
        """Test setting the state to the state mydevolo knows."""
        gateway = Gateway(self.gateway.get("id"), mydevolo_instance=mydevolo)
        gateway.online = False
        gateway.sync = False
        gateway.update_state()
        assert gateway.online
        assert gateway.sync



"""Exceptions occuring when talking to a Z-Wave device."""


class MprmDeviceCommunicationError(Exception):
    """Communicating to a device via mPRM failed."""


class MprmDeviceNotFoundError(Exception):
    """A device like this was not found."""


class WrongElementError(Exception):
    """This element was not meant for this property."""


class SwitchingProtected(Exception):
    """This device is protected against remote switching."""



import pytest
from devolo_home_control_api.exceptions.device import WrongElementError
from devolo_home_control_api.properties.remote_control_property import RemoteControlProperty


@pytest.mark.usefixtures("home_control_instance")
class TestRemoteControlProperty:
    def test_remote_control_property_invalid(self):
        with pytest.raises(WrongElementError):
            RemoteControlProperty(element_uid="invalid", setter=lambda uid, state: None)

    def test_set_invalid(self):
        with pytest.raises(ValueError):
            self.homecontrol.devices.get(self.devices.get("remote").get("uid")).remote_control_property.get(
                self.devices.get("remote").get("elementUIDs")[0]
            ).set(5)

    def test_set_valid(self):
        self.homecontrol.devices.get(self.devices.get("remote").get("uid")).remote_control_property.get(
            self.devices.get("remote").get("elementUIDs")[0]
        ).set(1)
        assert (
            self.homecontrol.devices.get(self.devices.get("remote").get("uid"))
            .remote_control_property.get(self.devices.get("remote").get("elementUIDs")[0])
            .key_pressed
        )



"""The Updater"""
import json
import logging
from contextlib import suppress
from typing import Any, Callable, Dict, List, Optional, Tuple

from ..backend import MESSAGE_TYPES
from ..devices.gateway import Gateway
from ..devices.zwave import Zwave
from ..helper.string import camel_case_to_snake_case
from ..helper.uid import get_device_type_from_element_uid, get_device_uid_from_element_uid, get_device_uid_from_setting_uid
from .publisher import Publisher


class Updater:  # pylint: disable=too-few-public-methods
    """
    The Updater takes care of new states and values of devices and sends them to the Publisher object. Using methods in here
    do not effect the real device states.

    :param devices: List of devices to await updates for
    :param gateway: Instance of a Gateway object
    :param publisher: Instance of a Publisher object
    """

    def __init__(self, devices: Dict[str, Zwave], gateway: Gateway, publisher: Publisher) -> None:
        self._logger = logging.getLogger(self.__class__.__name__)
        self._gateway = gateway
        self._publisher = publisher

        self.devices = devices
        self.on_device_change: Optional[Callable[[List[str]], Tuple[str, str]]] = None

    def update(self, message: Dict[str, Any]) -> None:
        """
        Update states and values depending on the message type.

        :param message: Message to process
        """
        unwanted_properties = [
            ".unregistering",
            "assistantsConnected",
            "operationStatus",
        ]

        # Early return on unwanted messages
        if (
            "UNREGISTERED" in message["topic"]
            or message["properties"]["property.name"] in unwanted_properties
            or "smartGroup" in message["properties"]["uid"]
        ):
            return

        # Handle pending operations messages
        if "property.name" in message["properties"] and message["properties"]["property.name"] == "pendingOperations":
            self._pending_operations(message)
            return

        # Handle all other messages
        message_type = MESSAGE_TYPES.get(get_device_type_from_element_uid(message["properties"]["uid"]), "_unknown")
        with suppress(AttributeError, KeyError):  # Sometime we receive already messages although the device is not setup yet.
            getattr(self, message_type)(message)

    def _automatic_calibration(self, message: Dict[str, Any]) -> None:
        """Update a automatic calibration message."""
        try:
            calibration_status = message["properties"]["property.value.new"]["status"]
            self._update_automatic_calibration(
                element_uid=message["properties"]["uid"], calibration_status=calibration_status != 2
            )
        except (KeyError, TypeError):
            if type(message["properties"]["property.value.new"]) not in [dict, list]:
                self._update_automatic_calibration(
                    element_uid=message["properties"]["uid"],
                    calibration_status=bool(message["properties"]["property.value.new"]),
                )

    def _binary_async(self, message: Dict[str, Any]) -> None:
        """Update a binary async setting."""
        if type(message["properties"]["property.value.new"]) not in [dict, list]:
            element_uid: str = message["properties"]["uid"]
            value = bool(message["properties"]["property.value.new"])
            device_uid = get_device_uid_from_setting_uid(element_uid)
            try:
                self.devices[device_uid].settings_property[camel_case_to_snake_case(element_uid).split("#")[-1]].value = value
            except KeyError:
                # Siren setting is not initialized like others.
                self.devices[device_uid].settings_property["muted"].value = value
            self._logger.debug("Updating state of %s to %s", element_uid, value)
            self._publisher.dispatch(device_uid, (element_uid, value))

    def _binary_sync(self, message: Dict[str, Any]) -> None:
        """Update a binary sync setting."""
        element_uid: str = message["properties"]["uid"]
        value = bool(message["properties"]["property.value.new"])
        device_uid = get_device_uid_from_setting_uid(element_uid)
        self.devices[device_uid].settings_property["movement_direction"].direction = value
        self._logger.debug("Updating state of %s to %s", element_uid, value)
        self._publisher.dispatch(device_uid, (element_uid, value))

    def _binary_sensor(self, message: Dict[str, Any]) -> None:
        """Update a binary sensor's state."""
        if message["properties"]["property.value.new"] is not None:
            element_uid: str = message["properties"]["uid"]
            value = bool(message["properties"]["property.value.new"])
            device_uid = get_device_uid_from_element_uid(element_uid)
            self.devices[device_uid].binary_sensor_property[element_uid].state = value
            self._logger.debug("Updating state of %s to %s", element_uid, value)
            self._publisher.dispatch(device_uid, (element_uid, value))

    def _binary_switch(self, message: Dict[str, Any]) -> None:
        """Update a binary switch's state."""
        if message["properties"]["property.name"] == "targetState" and message["properties"]["property.value.new"] is not None:
            element_uid: str = message["properties"]["uid"]
            value = bool(message["properties"]["property.value.new"])
            device_uid = get_device_uid_from_element_uid(element_uid)
            self.devices[device_uid].binary_switch_property[element_uid].state = value
            self._logger.debug("Updating state of %s to %s", element_uid, value)
            self._publisher.dispatch(device_uid, (element_uid, value))

    def _pending_operations(self, message: Dict[str, Any]) -> None:
        """Update pending operation state."""
        element_uid: str = message["properties"]["uid"]

        # Early return on useless messages
        if [
            uid
            for uid in ["devolo.HttpRequest", "devolo.PairDevice", "devolo.RemoveDevice", "devolo.mprm.gw.GatewayManager"]
            if (uid in element_uid)
        ]:
            return

        pending_operations = bool(message["properties"].get("property.value.new"))
        try:
            device_uid = get_device_uid_from_element_uid(element_uid)
            self.devices[device_uid].pending_operations = pending_operations
        except KeyError:
            device_uid = get_device_uid_from_setting_uid(element_uid)
            self.devices[device_uid].pending_operations = pending_operations
        self._logger.debug("Updating pending operations of device %s to %s", device_uid, pending_operations)
        self._publisher.dispatch(device_uid, ("pending_operations", pending_operations))

    def _current_consumption(self, message: Dict[str, Any]) -> None:
        """Update current consumption."""
        self._update_consumption(element_uid=message["uid"], consumption="current", value=message["property.value.new"])

    def _device_state(self, message: Dict[str, Any]) -> None:
        """Update the device state."""
        property_name = {
            "batteryLevel": "battery_level",
            "batteryLow": "battery_low",
            "status": "status",
        }

        device_uid = message["properties"]["uid"]
        name = message["properties"]["property.name"]
        value = message["properties"]["property.value.new"]

        try:
            self._logger.debug("Updating %s of %s to %s", property_name[name], device_uid, value)
            setattr(self.devices[device_uid], property_name[name], value)
            self._publisher.dispatch(device_uid, (device_uid, value, property_name[name]))
        except KeyError:
            self._unknown(message)

    def _gateway_accessible(self, message: Dict[str, Any]) -> None:
        """Update the gateway's state."""
        if message["properties"]["property.name"] == "gatewayAccessible":
            accessible = message["properties"]["property.value.new"]["accessible"]
            online_sync = message["properties"]["property.value.new"]["onlineSync"]
            self._logger.debug("Updating status and state of gateway to status: %s and state: %s", accessible, online_sync)
            self._gateway.online = accessible
            self._gateway.sync = online_sync

    def _general_device(self, message: Dict[str, Any]) -> None:
        """Update general device settings."""
        self._update_general_device_settings(
            element_uid=message["properties"]["uid"],
            events_enabled=message["properties"]["property.value.new"]["eventsEnabled"],
            icon=message["properties"]["property.value.new"]["icon"],
            name=message["properties"]["property.value.new"]["name"],
            zone_id=message["properties"]["property.value.new"]["zoneID"],
            zones=self._gateway.zones,
        )

    def _grouping(self, message: Dict[str, Any]) -> None:
        """Update zone (also called room) of a device."""
        self._gateway.zones = {key["id"]: key["name"] for key in message["properties"]["property.value.new"]}
        self._logger.debug("Updating gateway zones.")

    def _gui_enabled(self, message: Dict[str, Any]) -> None:
        """Update protection setting of binary switches."""
        device_uid = get_device_uid_from_element_uid(message["uid"])
        enabled = message["property.value.new"]
        for element_uid in self.devices[device_uid].binary_switch_property:
            self.devices[device_uid].binary_switch_property[element_uid].enabled = enabled
            self._logger.debug("Updating enabled state of %s to %s", element_uid, enabled)
            self._publisher.dispatch(device_uid, (element_uid, enabled, "gui_enabled"))

    def _humidity_bar(self, message: Dict[str, Any]) -> None:
        """Update a humidity bar."""
        fake_element_uid = f"devolo.HumidityBar:{message['properties']['uid'].split(':', 1)[1]}"
        value = message["properties"]["property.value.new"]
        device_uid = get_device_uid_from_element_uid(fake_element_uid)
        if message["properties"]["uid"].startswith("devolo.HumidityBarZone"):
            self.devices[device_uid].humidity_bar_property[fake_element_uid].zone = value
            self._logger.debug("Updating humidity bar zone of %s to %s", fake_element_uid, value)
        elif message["properties"]["uid"].startswith("devolo.HumidityBarValue"):
            self.devices[device_uid].humidity_bar_property[fake_element_uid].value = value
            self._logger.debug("Updating humidity bar value of %s to %s", fake_element_uid, value)
        self._publisher.dispatch(
            device_uid,
            (
                fake_element_uid,
                self.devices[device_uid].humidity_bar_property[fake_element_uid].zone,
                self.devices[device_uid].humidity_bar_property[fake_element_uid].value,
            ),
        )

    def _inspect_devices(self, message: Dict[str, Any]) -> None:
        """Call method if a new device appears or an old one disappears."""
        if not callable(self.on_device_change):
            self._logger.error("on_device_change is not set.")
            return

        if (
            not isinstance(message["properties"]["property.value.new"], list)
            or message["properties"]["uid"] != "devolo.DevicesPage"
        ):
            return

        device_uid, mode = self.on_device_change(message["properties"]["property.value.new"])
        if mode == "add":
            self._logger.info("%s added.", device_uid)
            self._publisher.add_event(event=device_uid)
            self._publisher.dispatch(device_uid, (device_uid, mode))
        else:
            self._publisher.dispatch(device_uid, (device_uid, mode))
            self._publisher.delete_event(event=device_uid)

    def _led(self, message: Dict[str, Any]) -> None:
        """Update LED settings."""
        if type(message["properties"]["property.value.new"]) not in [dict, list]:
            element_uid: str = message["properties"]["uid"]
            value = message["properties"]["property.value.new"]
            device_uid = get_device_uid_from_setting_uid(element_uid)
            self._logger.debug("Updating %s to %s.", element_uid, value)
            self.devices[device_uid].settings_property["led"].led_setting = value
            self._publisher.dispatch(device_uid, (element_uid, value))

    def _meter(self, message: Dict[str, Any]) -> None:
        """Update a meter value."""
        property_name = {
            "currentValue": self._current_consumption,
            "totalValue": self._total_consumption,
            "sinceTime": self._since_time,
            "guiEnabled": self._gui_enabled,
        }

        property_name[message["properties"]["property.name"]](message["properties"])

    def _multilevel_async(self, message: Dict[str, Any]) -> None:
        """Update multilevel async setting (mas) properties."""
        device_uid = get_device_uid_from_setting_uid(message["properties"]["uid"])
        try:
            name = camel_case_to_snake_case(message["properties"]["itemId"])
        # The Metering Plug has an multilevel async setting without an ID
        except KeyError:
            if self.devices[device_uid].device_model_uid == "devolo.model.Wall:Plug:Switch:and:Meter":
                name = "flash_mode"
            else:
                raise
        self.devices[device_uid].settings_property[name].value = message["properties"]["property.value.new"]

    def _multi_level_sensor(self, message: Dict[str, Any]) -> None:
        """Update a multi level sensor."""
        element_uid: str = message["properties"]["uid"]
        value = message["properties"]["property.value.new"]
        device_uid = get_device_uid_from_element_uid(element_uid)
        self._logger.debug("Updating %s to %s.", element_uid, value)
        self.devices[device_uid].multi_level_sensor_property[element_uid].value = value
        self._publisher.dispatch(device_uid, (element_uid, value))

    def _multi_level_switch(self, message: Dict[str, Any]) -> None:
        """Update a multi level switch."""
        if not isinstance(message["properties"]["property.value.new"], (list, dict, type(None))):
            element_uid: str = message["properties"]["uid"]
            value = message["properties"]["property.value.new"]
            device_uid = get_device_uid_from_element_uid(element_uid)
            self._logger.debug("Updating %s to %s.", element_uid, value)
            self.devices[device_uid].multi_level_switch_property[element_uid].value = value
            self._publisher.dispatch(device_uid, (element_uid, value))

    def _multilevel_sync(self, message: Dict[str, Any]) -> None:
        """Update multilevel sync settings."""
        if type(message["properties"]["property.value.new"]) not in [dict, list]:
            element_uid: str = message["properties"]["uid"]
            value = message["properties"]["property.value.new"]
            device_uid = get_device_uid_from_setting_uid(element_uid)
            device_model = self.devices[device_uid].device_model_uid
            self._logger.debug("Updating %s to %s.", element_uid, value)
            sync_type = {
                "devolo.model.Siren": "tone",
                "devolo.model.OldShutter": "shutter_duration",
                "devolo.model.Shutter": "shutter_duration",
            }

            try:
                setattr(self.devices[device_uid].settings_property[sync_type[device_model]], sync_type[device_model], value)
            except KeyError:
                # Other devices are up to now always motion sensors.
                self.devices[device_uid].settings_property["motion_sensitivity"].motion_sensitivity = value

            self._publisher.dispatch(device_uid, (element_uid, value))

    def _parameter(self, message: Dict[str, Any]) -> None:
        """Update parameter settings."""
        if type(message["properties"].get("property.value.new")) not in [dict, list]:
            element_uid: str = message["properties"]["uid"]
            param_changed = message["properties"]["property.value.new"]
            device_uid = get_device_uid_from_setting_uid(element_uid)
            self.devices[device_uid].settings_property["param_changed"].param_changed = param_changed
            self._logger.debug("Updating %s to %s.", element_uid, param_changed)
            self._publisher.dispatch(device_uid, (element_uid, param_changed))

    def _protection(self, message: Dict[str, Any]) -> None:
        """Update protection settings."""
        if type(message["properties"].get("property.value.new")) not in [dict, list]:
            element_uid: str = message["properties"]["uid"]
            value = message["properties"]["property.value.new"]
            name = message["properties"]["property.name"]
            device_uid = get_device_uid_from_setting_uid(element_uid)
            switching_type = {
                "targetLocalSwitch": "local_switching",
                "localSwitch": "local_switching",
                "targetRemoteSwitch": "remote_switching",
                "remoteSwitch": "remote_switching",
            }

            setattr(self.devices[device_uid].settings_property["protection"], switching_type[name], value)
            self._logger.debug("Updating %s protection of %s to %s", switching_type[name], element_uid, value)
            self._publisher.dispatch(device_uid, (element_uid, value, switching_type[name]))

    def _remote_control(self, message: Dict[str, Any]) -> None:
        """Update a remote control."""
        element_uid: str = message["properties"]["uid"]
        key_pressed = message["properties"]["property.value.new"]

        # The message for the diary needs to be ignored
        if key_pressed is not None:
            device_uid = get_device_uid_from_element_uid(element_uid)
            old_key_pressed = self.devices[device_uid].remote_control_property[element_uid].key_pressed
            self.devices[device_uid].remote_control_property[element_uid].key_pressed = key_pressed
            self._logger.debug(
                "Updating remote control of %s. Key %s",
                element_uid,
                f"pressed: {key_pressed}" if key_pressed != 0 else f"released: {old_key_pressed}",
            )
            self._publisher.dispatch(device_uid, (element_uid, key_pressed))

    def _since_time(self, message: Dict[str, Any]) -> None:
        """Update point in time the total consumption was reset."""
        element_uid = message["uid"]
        total_since = message["property.value.new"]
        device_uid = get_device_uid_from_element_uid(element_uid)
        self.devices[device_uid].consumption_property[element_uid].total_since = total_since
        self._logger.debug("Updating total since of %s to %s", element_uid, total_since)
        self._publisher.dispatch(device_uid, (element_uid, total_since, "total_since"))

    def _switch_type(self, message: Dict[str, Any]) -> None:
        """Update switch type setting (sts)."""
        element_uid: str = message["properties"]["uid"]
        value = message["properties"]["property.value.new"] * 2  # FWR, value.new is 1 for 2 buttons and 2 for 4 buttons.
        device_uid = get_device_uid_from_setting_uid(element_uid)
        self.devices[device_uid].settings_property["switch_type"].value = value
        self.devices[device_uid].remote_control_property[f"devolo.RemoteControl:{device_uid}"].key_count = value
        self._logger.debug("Updating switch type of %s to %s", device_uid, value)
        self._publisher.dispatch(device_uid, (element_uid, value))

    def _temperature_report(self, message: Dict[str, Any]) -> None:
        """Update temperature report settings."""
        if type(message["properties"].get("property.value.new")) not in [dict, list]:
            element_uid: str = message["properties"]["uid"]
            value = message["properties"]["property.value.new"]
            device_uid = get_device_uid_from_setting_uid(element_uid)
            self.devices[device_uid].settings_property["temperature_report"].temp_report = value
            self._logger.debug("Updating temperature report of %s to %s", element_uid, value)
            self._publisher.dispatch(device_uid, (element_uid, value))

    def _total_consumption(self, message: Dict[str, Any]) -> None:
        """Update total consumption."""
        self._update_consumption(element_uid=message["uid"], consumption="total", value=message["property.value.new"])

    def _unknown(self, message: Dict[str, Any]) -> None:
        """Ignore unknown messages."""
        ignore = (
            "devolo.DeviceEvents",
            "devolo.PairDevice",
            "devolo.SirenBinarySensor",
            "devolo.SirenMultiLevelSensor",
            "devolo.mprm.gw.GatewayManager",
            "devolo.mprm.gw.PortalManager",
            "ss",
            "mcs",
        )
        if not message["properties"]["uid"].startswith(ignore):
            self._logger.debug(json.dumps(message, indent=4))

    def _update_automatic_calibration(self, element_uid: str, calibration_status: bool) -> None:
        """Update automatic calibration setting of a device."""
        device_uid = get_device_uid_from_setting_uid(element_uid)
        self.devices[device_uid].settings_property["automatic_calibration"].calibration_status = calibration_status
        self._logger.debug("Updating value of %s to %s", element_uid, calibration_status)
        self._publisher.dispatch(device_uid, (element_uid, calibration_status))

    def _update_consumption(self, element_uid: str, consumption: str, value: float) -> None:
        """Update the consumption of a device."""
        device_uid = get_device_uid_from_element_uid(element_uid)
        setattr(self.devices[device_uid].consumption_property[element_uid], consumption, value)
        self._logger.debug("Updating %s consumption of %s to %s", consumption, element_uid, value)
        self._publisher.dispatch(device_uid, (element_uid, value, consumption))

    def _update_general_device_settings(self, element_uid: str, **kwargs: Any) -> None:
        """Update general device settings."""
        device_uid = get_device_uid_from_setting_uid(element_uid)
        for key, value in kwargs.items():
            setattr(self.devices[device_uid].settings_property["general_device_settings"], key, value)
            self._logger.debug("Updating attribute: %s of %s to %s", key, element_uid, value)
            self._publisher.dispatch(device_uid, (key, value))

    def _voltage_multi_level_sensor(self, message: Dict[str, Any]) -> None:
        """Update a voltage value."""
        element_uid: str = message["properties"]["uid"]
        value = message["properties"]["property.value.new"]
        device_uid = get_device_uid_from_element_uid(element_uid)
        self.devices[device_uid].multi_level_sensor_property[element_uid].value = value
        self._logger.debug("Updating voltage of %s to %s", element_uid, value)
        self._publisher.dispatch(device_uid, (element_uid, value))



from datetime import datetime

import pytest
from devolo_home_control_api.backend import MESSAGE_TYPES


@pytest.mark.usefixtures("mock_inspect_devices_metering_plug")
@pytest.mark.usefixtures("home_control_instance")
@pytest.mark.usefixtures("mock_mydevolo__call")
class TestHomeControl:
    def test_hasattr(self):
        homecontrol_functions = (
            func for func in dir(self.homecontrol) if callable(getattr(self.homecontrol, func)) and not func.startswith("__")
        )
        assert set(MESSAGE_TYPES.values()).difference(homecontrol_functions) == {"_gateway_accessible", "_device_state"}

    def test_binary_sensor_devices(self):
        assert hasattr(self.homecontrol.binary_sensor_devices[0], "binary_sensor_property")

    def test_binary_switch_devices(self):
        assert hasattr(self.homecontrol.binary_switch_devices[0], "binary_switch_property")

    def test_blinds_devices(self):
        assert hasattr(self.homecontrol.blinds_devices[0], "multi_level_switch_property")

    def test_multi_level_sensor_devices(self):
        assert hasattr(self.homecontrol.multi_level_sensor_devices[0], "multi_level_sensor_property")

    def test_multi_level_switch_devices(self):
        assert hasattr(self.homecontrol.multi_level_switch_devices[0], "multi_level_switch_property")

    def test_remote_control_devices(self):
        assert hasattr(self.homecontrol.remote_control_devices[0], "remote_control_property")

    def test_get_publisher(self):
        assert len(self.homecontrol.publisher._events) == 10

    def test__automatic_calibration(self):
        device = self.devices["blinds"]
        uid = device["uid"]
        self.homecontrol._automatic_calibration(
            {"UID": f"acs.{uid}", "properties": {"calibrationStatus": device["calibrationStatus"]}}
        )
        assert self.homecontrol.devices.get(uid).settings_property["automatic_calibration"].calibration_status == bool(
            device["calibrationStatus"]
        )

    def test___binary_sync(self):
        device = self.devices["blinds"]
        uid = device["uid"]
        self.homecontrol._binary_sync({"UID": f"acs.{uid}", "properties": {"value": device["inverted"]}})
        assert self.homecontrol.devices.get(uid).settings_property["movement_direction"].inverted is device["inverted"]

    def test__binary_async_blinds(self):
        device = self.devices.get("blinds").get("uid")
        i2 = self.devices.get("blinds").get("i2")
        self.homecontrol._binary_async({"UID": f"bas.{device}#i2", "properties": {"value": not i2}})
        assert self.homecontrol.devices.get(device).settings_property.get("i2").value is not i2

    def test__binary_async_siren(self):
        device = self.devices.get("siren").get("uid")
        muted = self.devices.get("siren").get("muted")
        self.homecontrol._binary_async({"UID": f"bas.{device}", "properties": {"value": not muted}})
        assert self.homecontrol.devices.get(device).settings_property.get("muted").value is not muted

    def test__binary_sensor(self):
        device = self.devices.get("sensor").get("uid")
        del self.homecontrol.devices[device].binary_sensor_property
        assert not hasattr(self.homecontrol.devices.get(device), "binary_sensor_property")
        self.homecontrol._binary_sensor(
            {
                "UID": self.devices.get("sensor").get("elementUIDs")[0],
                "properties": {
                    "state": self.devices.get("sensor").get("state"),
                    "sensorType": self.devices.get("sensor").get("sensor_type"),
                    "subType": "",
                },
            }
        )
        assert hasattr(self.homecontrol.devices.get(device), "binary_sensor_property")

    def test__binary_switch(self):
        device = self.devices["mains"]["uid"]
        del self.homecontrol.devices[device].binary_switch_property
        assert not hasattr(self.homecontrol.devices[device], "binary_switch_property")
        self.homecontrol._binary_switch(
            {
                "UID": self.devices["mains"]["properties"]["elementUIDs"][1],
                "properties": self.devices["mains"]["properties"],
            }
        )
        assert hasattr(self.homecontrol.devices[device], "binary_switch_property")

    def test__consumption(self):
        # TODO: Use test data
        device = self.devices.get("mains").get("uid")
        del self.homecontrol.devices[device].consumption_property
        assert not hasattr(self.homecontrol.devices.get(device), "consumption_property")
        self.homecontrol._meter(
            {
                "UID": "devolo.Meter:hdm:ZWave:F6BF9812/2",
                "properties": {
                    "currentValue": self.devices.get("mains").get("current_consumption"),
                    "totalValue": self.devices.get("mains").get("total_consumption"),
                    "sinceTime": self.devices.get("mains").get("properties").get("total_consumption"),
                },
            }
        )
        assert hasattr(self.homecontrol.devices.get(device), "consumption_property")

    def test__general_device(self):
        device = self.devices["mains"]
        element_uid = f"gds.{device['uid']}"
        self.homecontrol.devices[device["uid"]].settings_property["general_device_settings"].events_enabled = not device[
            "properties"
        ]["eventsEnabled"]
        self.homecontrol._general_device(
            {
                "UID": element_uid,
                "properties": {
                    "settings": {
                        "eventsEnabled": device["properties"]["eventsEnabled"],
                        "name": device["properties"]["itemName"],
                        "zoneID": device["properties"]["zoneId"],
                        "icon": device["properties"]["icon"],
                    }
                },
            }
        )
        assert (
            self.homecontrol.devices[device["uid"]].settings_property["general_device_settings"].events_enabled
            == device["properties"]["eventsEnabled"]
        )

    def test__humidity_bar(self):
        # TODO: Use test data
        device = self.devices.get("humidity").get("uid")
        del self.homecontrol.devices[device].humidity_bar_property
        assert not hasattr(self.homecontrol.devices.get(device), "humidity_bar_property")
        self.homecontrol._humidity_bar(
            {
                "UID": f"devolo.HumidityBarValue:{device}",
                "properties": {
                    "sensorType": "humidityBarPos",
                    "value": 75,
                },
            }
        )
        self.homecontrol._humidity_bar(
            {
                "UID": f"devolo.HumidityBarZone:{device}",
                "properties": {
                    "sensorType": "humidityBarZone",
                    "value": 1,
                },
            }
        )
        assert self.homecontrol.devices.get(device).humidity_bar_property.get(f"devolo.HumidityBar:{device}").value == 75
        assert self.homecontrol.devices.get(device).humidity_bar_property.get(f"devolo.HumidityBar:{device}").zone == 1

    def test__last_activity_binary_sensor(self):
        device = self.devices["sensor"]["uid"]
        element_uids = self.devices["sensor"]["elementUIDs"]
        self.homecontrol._binary_sensor(
            {
                "UID": element_uids[0],
                "properties": {
                    "state": self.devices["sensor"]["state"],
                    "sensorType": self.devices["sensor"]["sensor_type"],
                    "subType": "",
                },
            }
        )
        self.homecontrol._last_activity(
            {
                "UID": element_uids[1],
                "properties": {
                    "lastActivityTime": self.devices["sensor"]["last_activity"],
                },
            }
        )
        assert self.homecontrol.devices[device].binary_sensor_property.get(
            element_uids[0]
        ).last_activity == datetime.utcfromtimestamp(self.devices["sensor"]["last_activity"] / 1000)

    def test__last_activity_siren(self):
        device = self.devices["siren"]["uid"]
        element_uids = self.devices["siren"]["elementUIDs"]
        self.homecontrol._multi_level_switch(
            {
                "UID": element_uids[0],
                "properties": {
                    "value": self.devices["siren"]["properties"]["value"],
                    "switchType": self.devices["siren"]["switch_type"],
                    "max": self.devices["siren"]["properties"]["max"],
                    "min": self.devices["siren"]["properties"]["min"],
                },
            }
        )
        self.homecontrol._last_activity(
            {
                "UID": element_uids[3],
                "properties": {
                    "lastActivityTime": self.devices["siren"]["last_activity"],
                },
            }
        )
        assert self.homecontrol.devices[device].multi_level_switch_property[
            element_uids[0]
        ].last_activity == datetime.utcfromtimestamp(self.devices["siren"]["last_activity"] / 1000)

    def test__led(self):
        device = self.devices["mains"]
        element_uid = f"lis.{device['UID']}"
        led_setting = device["properties"]["led_setting"]
        self.homecontrol._led(
            {
                "UID": element_uid,
                "properties": {
                    "led": led_setting,
                },
            }
        )
        assert self.homecontrol.devices[device["UID"]].settings_property["led"].led_setting == led_setting
        device = self.devices["sensor"]
        element_uid = f"vfs.{device['UID']}"
        led_setting = device["led_setting"]
        self.homecontrol._led(
            {
                "UID": element_uid,
                "properties": {
                    "feedback": led_setting,
                },
            }
        )
        assert self.homecontrol.devices[device["UID"]].settings_property["led"].led_setting == led_setting

    def test__multilevel_async(self):
        device = self.devices["blinds"]
        uid = device["uid"]
        self.homecontrol._multilevel_async(
            {
                "UID": f"mas.{uid}",
                "properties": {
                    "itemId": "motorActivity",
                    "value": device["motorActivity"],
                },
            }
        )
        assert self.homecontrol.devices[uid].settings_property["motor_activity"].value == device["motorActivity"]

    def test__multilevel_async_mains(self):
        device = self.devices["mains"]
        uid = device["uid"]
        self.homecontrol._multilevel_async(
            {
                "UID": f"mas.{uid}",
                "properties": {
                    "itemId": None,
                    "value": device["flashMode"],
                },
            }
        )
        assert self.homecontrol.devices[uid].settings_property["flash_mode"].value == device["flashMode"]

    def test__multilevel_async_type_error(self):
        with pytest.raises(TypeError):
            device = self.devices["sensor"]
            uid = device["uid"]
            self.homecontrol._multilevel_async(
                {
                    "UID": f"mas.{uid}",
                    "properties": {
                        "itemId": None,
                        "value": device["motion_sensitivity"],
                    },
                }
            )

    def test__multilevel_sync_sensor(self):
        device = self.devices["sensor"]["uid"]
        self.homecontrol._multilevel_sync(
            {
                "UID": self.devices["sensor"]["settingUIDs"][2],
                "properties": self.devices["sensor"]["properties"],
            }
        )
        assert (
            self.homecontrol.devices[device].settings_property["motion_sensitivity"].motion_sensitivity
            == self.devices["sensor"]["properties"]["value"]
        )

    def test__multilevel_sync_shutter(self):
        device = self.devices["blinds"]["uid"]
        self.homecontrol._multilevel_sync(
            {
                "UID": f"mss.{device}",
                "properties": {
                    "value": self.devices["blinds"]["shutter_duration"],
                },
            }
        )
        assert (
            self.homecontrol.devices[device].settings_property["shutter_duration"].shutter_duration
            == self.devices["blinds"]["shutter_duration"]
        )

    def test__multilevel_sync_siren(self):
        device = self.devices["siren"]["uid"]
        self.homecontrol._multilevel_sync(
            {
                "UID": self.devices["siren"]["settingUIDs"][0],
                "properties": self.devices["siren"]["properties"],
            }
        )
        assert self.homecontrol.devices[device].settings_property["tone"].tone == self.devices["siren"]["properties"]["value"]

    def test__multi_level_sensor(self):
        # TODO: Use test data
        device = self.devices.get("sensor").get("uid")
        del self.homecontrol.devices[device].multi_level_sensor_property
        assert not hasattr(self.homecontrol.devices.get(device), "multi_level_sensor_property")
        self.homecontrol._multi_level_sensor(
            {
                "UID": self.devices.get("sensor").get("elementUIDs")[2],
                "properties": {
                    "value": 90.0,
                    "unit": "%",
                    "sensorType": "light",
                },
            }
        )
        assert hasattr(self.homecontrol.devices.get(device), "multi_level_sensor_property")

    def test__multi_level_switch(self):
        device = self.devices.get("siren").get("uid")
        del self.homecontrol.devices[device].multi_level_switch_property
        assert not hasattr(self.homecontrol.devices.get(device), "multi_level_switch_property")
        self.homecontrol._multi_level_switch(
            {
                "UID": self.devices.get("siren").get("elementUIDs")[0],
                "properties": {
                    "state": self.devices.get("multi_level_switch").get("state"),
                    "value": self.devices.get("multi_level_switch").get("value"),
                    "switchType": self.devices.get("multi_level_switch").get("switch_type"),
                    "max": self.devices.get("multi_level_switch").get("max"),
                    "min": self.devices.get("multi_level_switch").get("min"),
                },
            }
        )
        assert hasattr(self.homecontrol.devices.get(device), "multi_level_switch_property")

    def test__parameter(self):
        # TODO: Use test data
        device = self.devices.get("mains").get("uid")
        self.homecontrol._parameter(
            {
                "UID": "cps.hdm:ZWave:F6BF9812/2",
                "properties": {
                    "paramChanged": False,
                },
            }
        )
        assert hasattr(self.homecontrol.devices.get(device).settings_property.get("param_changed"), "param_changed")

    def test__protection(self):
        # TODO: Use test data
        device = self.devices.get("mains").get("uid")
        self.homecontrol._protection(
            {
                "UID": "ps.hdm:ZWave:F6BF9812/2",
                "properties": {
                    "localSwitch": True,
                    "remoteSwitch": False,
                },
            }
        )
        assert hasattr(self.homecontrol.devices.get(device).settings_property.get("protection"), "local_switching")
        assert hasattr(self.homecontrol.devices.get(device).settings_property.get("protection"), "remote_switching")

    def test__remote_control(self):
        device = self.devices.get("remote").get("uid")
        element_uid = self.devices.get("remote").get("elementUIDs")[0]
        del self.homecontrol.devices[device].remote_control_property
        assert not hasattr(self.homecontrol.devices.get(device), "remote_control_property")
        self.homecontrol._remote_control(
            {
                "UID": element_uid,
                "properties": {
                    "keyCount": self.devices.get("remote").get("key_count"),
                    "keyPressed": 0,
                    "type": 1,
                },
            }
        )
        assert self.homecontrol.devices.get(device).remote_control_property[element_uid].key_pressed == 0

    def test__switch_type(self):
        device = self.devices["remote"]["uid"]
        self.homecontrol._switch_type(
            {
                "UID": f"sts.{device}",
                "properties": {
                    "switchType": self.devices["remote"]["key_count"] / 2,
                },
            }
        )
        assert self.homecontrol.devices[device].settings_property["switch_type"].value == self.devices["remote"]["key_count"]

    def test__temperature_report(self):
        # TODO: Use test data
        device = self.devices.get("sensor").get("uid")
        self.homecontrol._temperature_report(
            {
                "UID": "trs.hdm:ZWave:F6BF9812/6",
                "properties": {
                    "tempReport": True,
                    "targetTempReport": False,
                },
            }
        )
        assert hasattr(self.homecontrol.devices.get(device).settings_property.get("temperature_report"), "temp_report")
        assert hasattr(self.homecontrol.devices.get(device).settings_property.get("temperature_report"), "target_temp_report")

    @pytest.mark.usefixtures("mock_inspect_devices")
    def test_device_change_add(self, mocker):
        uids = [self.devices.get(device).get("uid") for device in self.devices]
        uids.append("test_uid")
        spy = mocker.spy(self.homecontrol, "_inspect_devices")
        self.homecontrol.device_change(uids)
        spy.assert_called_once_with(["test_uid"])

    def test_device_change_remove(self):
        uids = [self.devices.get(device).get("uid") for device in self.devices]
        del uids[4]
        self.homecontrol.device_change(uids)
        assert self.devices.get("mains").get("uid") not in self.homecontrol.devices.keys()

    @pytest.mark.usefixtures("mock_extract_data_from_element_uids")
    @pytest.mark.usefixtures("mock_mprmrest_get_all_devices")
    def test__inspect_devices(self):
        uid = self.devices.get("mains").get("uid")
        del self.homecontrol.devices[uid]
        self.homecontrol._inspect_devices(self.homecontrol.get_all_devices())
        try:
            self.homecontrol.devices[uid]
            assert True
        except KeyError:
            assert False



import json
import pathlib

from devolo_home_control_api.devices.zwave import Zwave
from devolo_home_control_api.mydevolo import Mydevolo
from devolo_home_control_api.properties.binary_switch_property import BinarySwitchProperty
from devolo_home_control_api.properties.consumption_property import ConsumptionProperty
from devolo_home_control_api.properties.multi_level_sensor_property import MultiLevelSensorProperty
from devolo_home_control_api.properties.settings_property import SettingsProperty


def metering_plug(device_uid: str) -> Zwave:
    """
    Represent a metering plug in tests

    :param device_uid: Device UID this mock shall have
    :return: Metering Plug device
    """
    file = pathlib.Path(__file__).parent / ".." / "test_data.json"
    with file.open("r") as fh:
        test_data = json.load(fh)

    mydevolo = Mydevolo()
    device = Zwave(mydevolo_instance=mydevolo, **test_data["devices"]["mains"]["properties"])

    device.binary_switch_property = {}
    device.consumption_property = {}
    device.multi_level_sensor_property = {}
    device.settings_property = {}

    device.binary_switch_property[f"devolo.BinarySwitch:{device_uid}"] = BinarySwitchProperty(
        element_uid=f"devolo.BinarySwitch:{device_uid}",
        setter=lambda uid, state: None,
        state=test_data["devices"]["mains"]["properties"]["state"],
        enabled=test_data["devices"]["mains"]["properties"]["guiEnabled"],
    )
    device.consumption_property[f"devolo.Meter:{device_uid}"] = ConsumptionProperty(
        element_uid=f"devolo.Meter:{device_uid}",
        setter=lambda uid, state: None,
        current=test_data["devices"]["mains"]["properties"]["current_consumption"],
        total=test_data["devices"]["mains"]["properties"]["total_consumption"],
        total_since=test_data["devices"]["mains"]["properties"]["sinceTime"],
    )
    device.multi_level_sensor_property[f"devolo.VoltageMultiLevelSensor:{device_uid}"] = MultiLevelSensorProperty(
        element_uid=f"devolo.VoltageMultiLevelSensor:{device_uid}",
        current=test_data["devices"]["mains"]["properties"]["voltage"],
    )
    device.settings_property["param_changed"] = SettingsProperty(
        element_uid=f"cps.{device_uid}", setter=lambda uid, state: None
    )
    device.settings_property["general_device_settings"] = SettingsProperty(
        element_uid=f"gds.{device_uid}",
        setter=lambda uid, state: True,
        events_enabled=test_data["devices"]["mains"]["properties"]["eventsEnabled"],
        icon=test_data["devices"]["mains"]["properties"]["icon"],
        name=test_data["devices"]["mains"]["properties"]["itemName"],
        zone_id=test_data["devices"]["mains"]["properties"]["zoneId"],
        zones=test_data.get("gateway").get("zones"),
    )
    device.settings_property["led"] = SettingsProperty(
        element_uid=f"lis.{device_uid}",
        setter=lambda uid, state: True,
        led_setting=test_data["devices"]["mains"]["properties"]["led_setting"],
    )
    device.settings_property["protection"] = SettingsProperty(
        element_uid=f"ps.{device_uid}",
        setter=lambda uid, state: True,
        local_switching=test_data["devices"]["mains"]["properties"]["local_switch"],
        remote_switching=test_data["devices"]["mains"]["properties"]["remote_switch"],
    )
    device.settings_property["flash_mode"] = SettingsProperty(
        element_uid=f"mas.{device_uid}", setter=lambda uid, state: None, valus=test_data["devices"]["mains"]["flashMode"]
    )

    return device



from datetime import datetime, timezone

import pytest
from devolo_home_control_api.backend import MESSAGE_TYPES


@pytest.mark.usefixtures("home_control_instance")
@pytest.mark.usefixtures("mock_publisher_dispatch")
class TestUpdater:
    def test_hasattr(self):
        updater_functions = (
            func
            for func in dir(self.homecontrol.updater)
            if callable(getattr(self.homecontrol.updater, func)) and not func.startswith("__")
        )
        assert set(MESSAGE_TYPES.values()).difference(updater_functions) == set()

    @pytest.mark.usefixtures("mock_updater_binary_switch")
    def test_update_device(self, mocker):
        message = {
            "topic": "com/prosyst/mbs/services/fim/FunctionalItemEvent/PROPERTY_CHANGED",
            "properties": {
                "property.name": "state",
                "uid": f"devolo.BinarySwitch:{self.devices['mains']['uid']}",
            },
        }
        spy = mocker.spy(self.homecontrol.updater, "_binary_switch")
        self.homecontrol.updater.update(message=message)
        spy.assert_called_once_with(message)

    @pytest.mark.usefixtures("mock_updater_pending_operations")
    def test_update_pending_operations(self, mocker):
        message = {
            "topic": "com/prosyst/mbs/services/fim/FunctionalItemEvent/PROPERTY_CHANGED",
            "properties": {
                "property.name": "pendingOperations",
                "uid": "",
            },
        }
        spy = mocker.spy(self.homecontrol.updater, "_pending_operations")
        self.homecontrol.updater.update(message=message)
        spy.assert_called_once_with(message)

    def test_update_unwanted(self, mocker):
        message = {
            "topic": "com/prosyst/mbs/services/fim/FunctionalItemEvent/UNREGISTERED",
        }
        spy = mocker.spy(self.homecontrol.updater, "_unknown")
        self.homecontrol.updater.update(message=message)
        spy.assert_not_called()

    def test_update_smart_group(self, mocker):
        message = {
            "topic": "com/prosyst/mbs/services/fim/FunctionalItemEvent/PROPERTY_CHANGED",
            "properties": {
                "uid": "devolo.MultiLevelSwitch:devolo.smartGroup.Group",
                "property.name": "value",
            },
        }
        spy = mocker.spy(self.homecontrol.updater, "_multi_level_switch")
        self.homecontrol.updater.update(message=message)
        spy.assert_not_called()
        message = {
            "topic": "com/prosyst/mbs/services/fim/FunctionalItemEvent/PROPERTY_CHANGED",
            "properties": {
                "uid": "devolo.BinarySwitch:devolo.smartGroup.Group",
                "property.name": "value",
            },
        }
        spy = mocker.spy(self.homecontrol.updater, "_binary_switch")
        self.homecontrol.updater.update(message=message)
        spy.assert_not_called()

    def test__automatic_calibration(self):
        uid = self.devices["blinds"]["uid"]
        calibration_status = self.devices["blinds"]["calibrationStatus"]
        self.homecontrol.updater._automatic_calibration(
            message={
                "properties": {
                    "uid": f"acs.{uid}",
                    "property.value.new": {
                        "status": calibration_status,
                    },
                }
            }
        )
        assert self.homecontrol.devices[uid].settings_property["automatic_calibration"].calibration_status

    def test__automatic_calibration_key_error(self):
        uid = self.devices["blinds"]["uid"]
        calibration_status = self.devices["blinds"]["calibrationStatus"]
        self.homecontrol.updater._automatic_calibration(
            message={
                "properties": {
                    "uid": f"acs.{uid}",
                    "property.value.new": calibration_status,
                }
            }
        )
        assert self.homecontrol.devices[uid].settings_property["automatic_calibration"].calibration_status

    def test__binary_async_blinds(self):
        uid = self.devices["blinds"]["uid"]
        self.homecontrol.devices[uid].settings_property["i2"].value = self.devices["blinds"]["i2"]
        self.homecontrol.updater._binary_async(
            message={
                "properties": {
                    "uid": f"bas.{uid}#i2",
                    "property.value.new": not self.devices["blinds"]["i2"],
                }
            }
        )
        assert not self.homecontrol.devices[uid].settings_property["i2"].value

    def test__binary_async_siren(self):
        uid = self.devices["siren"]["uid"]
        self.homecontrol.devices[uid].settings_property["muted"].value = self.devices["siren"]["muted"]
        self.homecontrol.updater._binary_async(
            message={
                "properties": {
                    "uid": f"bas.{uid}",
                    "property.value.new": not self.devices["siren"]["muted"],
                }
            }
        )
        assert not self.homecontrol.devices[uid].settings_property["muted"].value

    def test__binary_sensor(self):
        uid = self.devices["sensor"]["uid"]
        device = self.homecontrol.devices[uid].binary_sensor_property[f"devolo.BinarySensor:{uid}"]
        device.state = True
        state = device.state
        self.homecontrol.updater._binary_sensor(
            message={
                "properties": {
                    "property.name": "state",
                    "uid": f"devolo.BinarySensor:{uid}",
                    "property.value.new": 0,
                }
            }
        )
        state_new = device.state
        assert state != state_new
        assert device.last_activity != datetime.fromtimestamp(0)

    def test__binary_sync(self):
        uid = self.devices["blinds"]["uid"]
        self.homecontrol.updater._binary_sync(
            message={
                "properties": {
                    "uid": f"bss.{uid}",
                    "property.value.new": self.devices["blinds"]["movement_direction"],
                },
            }
        )
        assert self.homecontrol.devices[uid].settings_property["movement_direction"].direction is bool(
            self.devices["blinds"]["movement_direction"]
        )

    def test__binary_switch(self):
        uid = self.devices.get("mains").get("uid")
        self.homecontrol.devices.get(uid).binary_switch_property.get(f"devolo.BinarySwitch:{uid}").state = True
        state = self.homecontrol.devices.get(uid).binary_switch_property.get(f"devolo.BinarySwitch:{uid}").state
        self.homecontrol.updater._binary_switch(
            message={
                "properties": {"property.name": "targetState", "uid": f"devolo.BinarySwitch:{uid}", "property.value.new": 0},
            }
        )
        state_new = self.homecontrol.devices.get(uid).binary_switch_property.get(f"devolo.BinarySwitch:{uid}").state
        assert state != state_new

    def test__device_state(self):
        uid = self.devices.get("mains").get("uid")
        online_state = self.homecontrol.devices.get(uid).status
        self.homecontrol.updater._device_state(
            message={"properties": {"uid": uid, "property.name": "status", "property.value.new": 1}}
        )
        assert self.homecontrol.devices.get(uid).status == 1
        assert online_state != self.homecontrol.devices.get(uid).status

    def test__general_device(self):
        uid = self.devices["mains"]["uid"]
        events_enabled = self.devices["mains"]["properties"]["eventsEnabled"]
        self.homecontrol.devices[uid].settings_property["general_device_settings"].events_enabled = events_enabled
        self.homecontrol.updater._general_device(
            message={
                "properties": {
                    "uid": f"gds.{uid}",
                    "property.value.new": {
                        "eventsEnabled": not events_enabled,
                        "icon": self.devices["mains"]["properties"]["icon"],
                        "name": self.devices["mains"]["properties"]["itemName"],
                        "zoneID": self.devices["mains"]["properties"]["zoneId"],
                    },
                },
            }
        )
        assert not self.homecontrol.devices[uid].settings_property["general_device_settings"].events_enabled

    def test__gateway_accessible(self):
        self.homecontrol.gateway.online = True
        self.homecontrol.gateway.sync = True
        accessible = self.homecontrol.gateway.online
        online_sync = self.homecontrol.gateway.sync
        self.homecontrol.updater._gateway_accessible(
            message={
                "properties": {
                    "property.name": "gatewayAccessible",
                    "property.value.new": {
                        "accessible": False,
                        "onlineSync": False,
                    },
                },
            }
        )
        accessible_new = self.homecontrol.gateway.online
        online_sync_new = self.homecontrol.gateway.sync
        assert accessible != accessible_new
        assert online_sync != online_sync_new

    def test__grouping(self):
        zone_id = "hz_3"
        name = self.gateway["zones"][zone_id]
        self.homecontrol.updater._grouping(
            message={
                "properties": {
                    "property.value.new": [
                        {
                            "id": zone_id,
                            "name": self.devices["mains"]["properties"]["zone"],
                        }
                    ],
                },
            }
        )
        assert self.homecontrol.gateway.zones[zone_id] != name
        assert self.homecontrol.gateway.zones[zone_id] == self.devices["mains"]["properties"]["zone"]

    def test__gui_enabled(self):
        uid = self.devices["mains"]["uid"]
        element_uids = self.devices["mains"]["properties"]["elementUIDs"]
        enabled = self.devices["mains"]["properties"]["guiEnabled"]
        self.homecontrol.devices[uid].binary_switch_property[element_uids[1]].enabled = enabled
        self.homecontrol.updater._gui_enabled(
            message={
                "uid": element_uids[0],
                "property.value.new": not enabled,
            }
        )
        assert self.homecontrol.devices[uid].binary_switch_property[element_uids[1]].enabled != enabled

    def test__humidity_bar(self):
        uid = self.devices.get("humidity").get("uid")
        self.homecontrol.devices.get(uid).humidity_bar_property.get(f"devolo.HumidityBar:{uid}").value = 75
        self.homecontrol.devices.get(uid).humidity_bar_property.get(f"devolo.HumidityBar:{uid}").zone = 1
        current_value = self.homecontrol.devices.get(uid).humidity_bar_property.get(f"devolo.HumidityBar:{uid}").value
        current_zone = self.homecontrol.devices.get(uid).humidity_bar_property.get(f"devolo.HumidityBar:{uid}").zone
        self.homecontrol.updater._humidity_bar(
            message={
                "properties": {
                    "uid": f"devolo.HumidityBarValue:{uid}",
                    "property.value.new": 50,
                }
            }
        )
        current_value_new = self.homecontrol.devices.get(uid).humidity_bar_property.get(f"devolo.HumidityBar:{uid}").value
        self.homecontrol.updater._humidity_bar(
            message={
                "properties": {
                    "uid": f"devolo.HumidityBarZone:{uid}",
                    "property.value.new": 0,
                }
            }
        )
        current_zone_new = self.homecontrol.devices.get(uid).humidity_bar_property.get(f"devolo.HumidityBar:{uid}").zone

        assert current_value_new == 50
        assert current_zone_new == 0
        assert current_value != current_value_new
        assert current_zone != current_zone_new

    @pytest.mark.usefixtures("mock_publisher_add_event")
    @pytest.mark.usefixtures("mock_publisher_delete_event")
    @pytest.mark.parametrize("event", ["add", "delete"])
    def test__inspect_devices_valid(self, mocker, event):
        self.homecontrol.updater.on_device_change = lambda device_uids: (device_uids, event)
        spy = mocker.spy(self.homecontrol.updater._publisher, f"{event}_event")
        self.homecontrol.updater._inspect_devices(
            message={
                "properties": {
                    "uid": "devolo.DevicesPage",
                    "property.value.new": [],
                }
            }
        )
        spy.assert_called_once()

    def test__inspect_devices_error(self, mocker):
        self.homecontrol.updater.on_device_change = None
        spy = mocker.spy(self.homecontrol.updater._logger, "error")
        self.homecontrol.updater._inspect_devices({})
        spy.assert_called_once_with("on_device_change is not set.")
        self.homecontrol.updater.on_device_change = lambda: None
        spy = mocker.spy(self.homecontrol.updater, "on_device_change")
        self.homecontrol.updater._inspect_devices(
            message={
                "properties": {
                    "uid": "",
                    "property.value.new": [],
                }
            }
        )
        spy.assert_not_called()

    def test__meter(self):
        uid = self.devices.get("mains").get("uid")
        self.homecontrol.devices.get(uid).consumption_property.get(f"devolo.Meter:{uid}").current = 5
        self.homecontrol.devices.get(uid).consumption_property.get(f"devolo.Meter:{uid}").total = 230
        total = self.homecontrol.devices.get(uid).consumption_property.get(f"devolo.Meter:{uid}").total
        # Changing current value
        self.homecontrol.updater._meter(
            message={
                "properties": {
                    "property.name": "currentValue",
                    "uid": f"devolo.Meter:{uid}",
                    "property.value.new": 7,
                }
            }
        )
        current_new = self.homecontrol.devices.get(uid).consumption_property.get(f"devolo.Meter:{uid}").current
        # Check if current value has changed
        assert current_new == 7
        # Check if total has not changed
        assert total == self.homecontrol.devices.get(uid).consumption_property.get(f"devolo.Meter:{uid}").total
        # Changing total value
        self.homecontrol.updater._meter(
            message={"properties": {"property.name": "totalValue", "uid": f"devolo.Meter:{uid}", "property.value.new": 235}}
        )
        total_new = self.homecontrol.devices.get(uid).consumption_property.get(f"devolo.Meter:{uid}").total
        # Check if total value has changed
        assert total_new == 235
        # Check if current value has not changed
        assert self.homecontrol.devices.get(uid).consumption_property.get(f"devolo.Meter:{uid}").current == current_new

    def test___multilevel_async(self):
        uid = self.devices["mains"]["UID"]
        element_uid = self.devices["mains"]["properties"]["settingUIDs"][4]
        value = self.devices["mains"]["flashMode"] + 1
        self.homecontrol.updater._multilevel_async(message={"properties": {"uid": element_uid, "property.value.new": value}})
        assert self.homecontrol.devices[uid].settings_property["flash_mode"].value == value

    def test__multi_level_sensor(self):
        uid = self.devices["sensor"]["uid"]
        element_uid = f"devolo.MultiLevelSensor:{uid}#MultilevelSensor(1)"
        self.homecontrol.devices[uid].multi_level_sensor_property[element_uid].value = 100
        current = self.homecontrol.devices[uid].multi_level_sensor_property[element_uid].value
        self.homecontrol.updater._multi_level_sensor(message={"properties": {"uid": element_uid, "property.value.new": 50}})
        current_new = self.homecontrol.devices[uid].multi_level_sensor_property[element_uid].value

        assert current_new == 50
        assert current != current_new

    def test__multi_level_switch(self):
        device = self.devices.get("multi_level_switch")
        uid = device.get("uid")
        self.homecontrol.devices.get(uid).multi_level_switch_property.get(device.get("elementUIDs")[0]).value = device.get(
            "value"
        )
        current = self.homecontrol.devices.get(uid).multi_level_switch_property.get(device.get("elementUIDs")[0]).value
        self.homecontrol.updater._multi_level_switch(
            message={
                "properties": {
                    "uid": device.get("elementUIDs")[0],
                    "property.value.new": device.get("max"),
                }
            }
        )
        current_new = self.homecontrol.devices.get(uid).multi_level_switch_property.get(device.get("elementUIDs")[0]).value

        assert current_new == device.get("max")
        assert current != current_new

    def test__multilevel_sync_sensor(self):
        device = self.devices["sensor"]
        uid = device["uid"]
        value = device["properties"]["value"]
        self.homecontrol.devices[uid].settings_property["motion_sensitivity"].motion_sensitivity = value
        self.homecontrol.updater._multilevel_sync(
            message={
                "properties": {
                    "uid": f"mss.{uid}",
                    "property.value.new": value - 1,
                }
            }
        )
        assert self.homecontrol.devices[uid].settings_property["motion_sensitivity"].motion_sensitivity == value - 1

    def test__multilevel_sync_shutter(self):
        device = self.devices["blinds"]
        uid = device["uid"]
        value = device["shutter_duration"]
        self.homecontrol.devices[uid].settings_property["shutter_duration"].shutter_duration = value
        self.homecontrol.updater._multilevel_sync(
            message={
                "properties": {
                    "uid": f"mss.{uid}",
                    "property.value.new": value - 1,
                }
            }
        )
        assert self.homecontrol.devices[uid].settings_property["shutter_duration"].shutter_duration == value - 1

    def test__multilevel_sync_siren(self):
        device = self.devices["siren"]
        uid = device["uid"]
        value = device["properties"]["value"]
        self.homecontrol.devices[uid].settings_property["tone"].tone = value
        self.homecontrol.updater._multilevel_sync(
            message={
                "properties": {
                    "uid": f"mss.{uid}",
                    "property.value.new": value - 1,
                }
            }
        )
        assert self.homecontrol.devices[uid].settings_property["tone"].tone == value - 1

    def test__led(self):
        device = self.devices["mains"]
        uid = device["uid"]
        led_setting = device["properties"]["led_setting"]
        self.homecontrol.devices[uid].settings_property["led"].led_setting = led_setting
        self.homecontrol.updater._led(
            message={
                "properties": {
                    "uid": f"lis.{uid}",
                    "property.value.new": not led_setting,
                }
            }
        )
        assert self.homecontrol.devices[uid].settings_property["led"].led_setting is not led_setting

    def test__parameter(self):
        device = self.devices["mains"]
        uid = device["uid"]
        param_changed = device["properties"]["param_changed"]
        self.homecontrol.devices[uid].settings_property["param_changed"].param_changed = param_changed
        self.homecontrol.updater._parameter(
            message={
                "properties": {
                    "uid": f"cps.{uid}",
                    "property.value.new": not param_changed,
                }
            }
        )
        assert self.homecontrol.devices[uid].settings_property["param_changed"].param_changed is not param_changed

    def test__pending_operations_false(self):
        device = self.devices["mains"]
        uid = device["uid"]
        pending_operation = device["properties"]["pending_operations"]
        self.homecontrol.devices[uid].pending_operation = not pending_operation
        self.homecontrol.updater._pending_operations(message={"properties": {"uid": device["elementUIDs"][1]}})
        assert self.homecontrol.devices[uid].pending_operation

    def test__pending_operations_true(self):
        device = self.devices["mains"]
        uid = device["uid"]
        pending_operation = device["properties"]["pending_operations"]
        self.homecontrol.devices[uid].pending_operation = pending_operation
        self.homecontrol.updater._pending_operations(
            message={"properties": {"uid": device["elementUIDs"][1], "property.value.new": {"status": 1}}}
        )
        assert not self.homecontrol.devices[uid].pending_operation

    def test__pending_operations_setting(self):
        device = self.devices["mains"]
        uid = device["uid"]
        pending_operation = device["properties"]["pending_operations"]
        self.homecontrol.devices[uid].pending_operation = pending_operation
        self.homecontrol.updater._pending_operations(
            message={
                "properties": {
                    "uid": device["settingUIDs"][3],
                    "property.value.new": {
                        "status": 1,
                    },
                }
            }
        )
        assert not self.homecontrol.devices[uid].pending_operation

    def test__pending_operations_useless(self, mocker):
        spy = mocker.spy(self.homecontrol.updater._publisher, "dispatch")
        for uid in ["devolo.HttpRequest", "devolo.PairDevice", "devolo.RemoveDevice", "devolo.mprm.gw.GatewayManager"]:
            self.homecontrol.updater._pending_operations(
                message={
                    "properties": {
                        "uid": uid,
                    }
                }
            )
            spy.assert_not_called()

    def test__protection_local(self):
        device = self.devices["mains"]
        uid = device["uid"]
        local_switch = device["properties"]["local_switch"]
        self.homecontrol.devices[uid].settings_property["protection"].local_switching = local_switch
        self.homecontrol.updater._protection(
            message={
                "properties": {
                    "uid": f"ps.{uid}",
                    "property.name": "targetLocalSwitch",
                    "property.value.new": not local_switch,
                },
            }
        )
        assert self.homecontrol.devices[uid].settings_property["protection"].local_switching is not local_switch

    def test__protection_remote(self):
        device = self.devices["mains"]
        uid = device["uid"]
        remote_switch = device["properties"]["remote_switch"]
        self.homecontrol.devices[uid].settings_property["protection"].remote_switching = remote_switch
        self.homecontrol.updater._protection(
            message={
                "properties": {
                    "uid": f"ps.{uid}",
                    "property.name": "targetRemoteSwitch",
                    "property.value.new": not remote_switch,
                },
            }
        )
        assert self.homecontrol.devices[uid].settings_property["protection"].remote_switching is not remote_switch

    def test__remote_control(self):
        device = self.devices.get("remote")
        uid = device.get("uid")
        self.homecontrol.devices.get(uid).remote_control_property.get(device.get("elementUIDs")[0]).key_pressed = 0
        self.homecontrol.updater._remote_control(
            message={
                "properties": {
                    "uid": device.get("elementUIDs")[0],
                    "property.value.new": 1,
                }
            }
        )

        assert self.homecontrol.devices.get(uid).remote_control_property.get(device.get("elementUIDs")[0]).key_pressed == 1

    def test__since_time(self):
        device = self.devices["mains"]
        uid = device["uid"]
        now = datetime.now()
        total_since = self.homecontrol.devices[uid].consumption_property[f"devolo.Meter:{uid}"].total_since
        self.homecontrol.updater._since_time(
            {
                "uid": f"devolo.Meter:{uid}",
                "property.value.new": now.replace(tzinfo=timezone.utc).timestamp() * 1000,
            }
        )
        new_total_since = self.homecontrol.devices[uid].consumption_property[f"devolo.Meter:{uid}"].total_since
        assert total_since != new_total_since
        assert new_total_since == now

    def test__switch_type(self):
        device = self.devices["remote"]
        uid = device["uid"]
        self.homecontrol.updater._switch_type(
            message={
                "properties": {
                    "uid": f"sts.{uid}",
                    "property.value.new": device["key_count"] / 4,
                }
            }
        )
        assert self.homecontrol.devices[uid].settings_property["switch_type"].value == device["key_count"] / 2

    def test__temperature_report(self):
        device = self.devices["sensor"]
        uid = device["uid"]
        self.homecontrol.devices[uid].settings_property["temperature_report"].temp_report = device["temp_report"]
        self.homecontrol.updater._temperature_report(
            message={
                "properties": {
                    "uid": f"trs.{uid}",
                    "property.value.new": not device["temp_report"],
                }
            }
        )
        assert self.homecontrol.devices[uid].settings_property["temperature_report"].temp_report is not device["temp_report"]

    def test__voltage_multi_level_sensor(self):
        uid = self.devices.get("mains").get("uid")
        self.homecontrol.devices.get(uid).multi_level_sensor_property.get(f"devolo.VoltageMultiLevelSensor:{uid}").value = 231
        current = (
            self.homecontrol.devices.get(uid).multi_level_sensor_property.get(f"devolo.VoltageMultiLevelSensor:{uid}").value
        )
        self.homecontrol.updater._voltage_multi_level_sensor(
            message={"properties": {"uid": f"devolo.VoltageMultiLevelSensor:{uid}", "property.value.new": 234}}
        )
        current_new = (
            self.homecontrol.devices.get(uid).multi_level_sensor_property.get(f"devolo.VoltageMultiLevelSensor:{uid}").value
        )

        assert current_new == 234
        assert current != current_new
